/*
 * ################################################################
 *
 * ProActive: The Java(TM) library for Parallel, Distributed,
 *            Concurrent computing with Security and Mobility
 *
 * Copyright (C) 1997-2002 INRIA/University of Nice-Sophia Antipolis
 * Contact: proactive-support@inria.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Initial developer(s):               The ProActive Team
 *                        http://www.inria.fr/oasis/ProActive/contacts.html
 *  Contributor(s):
 *
 * ################################################################
 */
package org.objectweb.proactive.core.node;

import java.net.UnknownHostException;

import org.apache.log4j.Logger;
import org.objectweb.proactive.ProActive;
import org.objectweb.proactive.core.ProActiveException;
import org.objectweb.proactive.core.UniqueID;
import org.objectweb.proactive.core.config.ProActiveConfiguration;
import org.objectweb.proactive.core.runtime.ProActiveRuntime;
import org.objectweb.proactive.core.runtime.RuntimeFactory;
import org.objectweb.proactive.core.util.UrlBuilder;
import org.objectweb.proactive.ext.security.PolicyServer;


/**
 * <p>
 * The <code>NodeFactory</code> provides a generic way to create and lookup <code>Node</code>
 * without protocol specific code (such as RMI or Jini).
 * </p><p>
 * <code>NodeFactory</code> provides a set of static methods to create and lookup <code>Node</code>.
 * To create a node it is only necessary to associate the protocol in the node url.
 * For instance :
 * </p>
 * <pre>
 *    rmi://localhost/node1
 *    jini://localhost/node2
 * </pre>
 * <p>
 * As long as a protocol specific factory has been registered for the
 * given protocol, the creation of the node will be delegated to the right factory.
 * </p><p>
 * This class also provide the concept of default node and default protocol. When the protocol is not
 * specified in the node URL, the default protocol is used. When an active object is created in the local
 * JVM but without being attached to any node, a default node is created to hold that active object.
 * </p>
 *
 * @author  ProActive Team
 * @version 1.1,  2002/08/28
 * @since   ProActive 0.9
 *
 */
public class NodeFactory {
    protected static Logger logger = Logger.getLogger(NodeFactory.class.getName());
    private static final String DEFAULT_NODE_NAME;
    private static Node defaultNode = null;

    static {
        ProActiveConfiguration.load();

        String protocol = UrlBuilder.checkProtocol(System.getProperty(
                    "proactive.communication.protocol"));
        String port = System.getProperty("proactive.rmi.port");
        if (port != null) {
            DEFAULT_NODE_NAME = UrlBuilder.buildUrl("localhost", "Node",
                    protocol, new Integer(port).intValue());
        } else {
            DEFAULT_NODE_NAME = UrlBuilder.buildUrl("localhost", "Node",
                    protocol);
        }
    }

    // test with class loader
    //private static final ClassLoader myClassLoader = new NodeClassLoader();
    //
    // -- PUBLIC METHODS - STATIC -----------------------------------------------
    //

    /**
     * Returns the reference to the default node associated to the current JVM
     * If no default node yet exists, it creates a new one.
     * @return a reference to the default node associated to this JVM
     * @exception NodeException if the default node cannot be instantiated
     */
    public static synchronized Node getDefaultNode() throws NodeException {
        String nodeURL;
        ProActiveRuntime defaultRuntime;
        String jobID = ProActive.getJobId();
        if (defaultNode == null) {
            try {
                defaultRuntime = RuntimeFactory.getDefaultRuntime();
                nodeURL = defaultRuntime.createLocalNode(DEFAULT_NODE_NAME +
                        Integer.toString(
                            new java.util.Random(System.currentTimeMillis()).nextInt()),
                        false,defaultRuntime.getPolicyServer(), "currentJVM",jobID );
            } catch (ProActiveException e) {
                throw new NodeException("Cannot create the default Node", e);
            }
            defaultNode = new NodeImpl(defaultRuntime, nodeURL,
                    UrlBuilder.checkProtocol(System.getProperty("proactive.communication.protocol")), jobID);
        }
        return defaultNode;
    }

    /**
     * Returns true if the given node belongs to this JVM false else.
     * @return true if the given node belongs to this JVM false else
     */
    public static boolean isNodeLocal(Node node) {
        return node.getNodeInformation().getVMID().equals(UniqueID.getCurrentVMID());
    }

    /**
     * Creates a new node on the local machine. This call can only be used
     * to create a node on the local JVM on the local machine.
     * The node URL can be in the form
     * <ul>
     * <li>///nodeName</li>
     * <li>//localhost/nodeName</li>
     * <li>//<i>&lt;hostname></i>/nodeName</li>
     * </ul>
     * where <i>&lt;hostname></i> is the name of the localhost.
     * @param <code>nodeURL</code> the URL of the node to create
     * @return the newly created node on the local JVM
     * @exception NodeException if the node cannot be created
     */
    public static Node createNode(String nodeURL) throws NodeException {
        return createNode(nodeURL, false,null,null);
    }

    /**
     * Creates a new node on the local machine. This call can only be used
     * to create a node on the local JVM on the local machine.
     * The node URL can be in the form
     * <ul>
     * <li>///nodeName</li>
     * <li>//localhost/nodeName</li>
     * <li>//<i>&lt;hostname></i>/nodeName</li>
     * </ul>
     * where <i>&lt;hostname></i> is the name of the localhost.
     * @param <code>nodeURL</code> the URL of the node to create
     * @param <code>replacePreviousBinding</code>
     * @return the newly created node on the local JVM
     * @exception NodeException if the node cannot be created
     */
    public static Node createNode(String url, boolean replacePreviousBinding, PolicyServer ps, String vnname)
        throws NodeException {
        ProActiveRuntime proActiveRuntime;
        String nodeURL;
        String jobID = ProActive.getJobId();

        if (logger.isDebugEnabled()) {
            logger.debug("NodeFactory: createNode(" + url + ")");
        }

        //first look for the prototcol
        String protocol = UrlBuilder.getProtocol(url);

        //NodeFactory factory = getFactory(protocol);
        //then create a node
        try {
            proActiveRuntime = RuntimeFactory.getProtocolSpecificRuntime(protocol);
            nodeURL = proActiveRuntime.createLocalNode(url,
                    replacePreviousBinding,ps,vnname, jobID);
        } catch (ProActiveException e) {
            throw new NodeException("Cannot create a Node based on " + url, e);
        }
        Node node = new NodeImpl(proActiveRuntime, nodeURL, protocol, jobID);
        return node;
    }

    /**
     * Returns the reference to the node located at the given url.
     * This url can be either local or remote.
     * @param nodeURL. The url of the node
     * @return Node. The reference of the node
     * @throws NodeException if the node cannot be found
     */
    public static Node getNode(String nodeURL) throws NodeException {
        ProActiveRuntime proActiveRuntime;
        String url;
        String nodeName;
        String jobID;

        if (logger.isDebugEnabled()) {
            logger.debug("NodeFactory: getNode() for " + nodeURL);
        }

        //do we have any association for this node?
        String protocol = UrlBuilder.getProtocol(nodeURL);

        //String noProtocolUrl = UrlBuilder.removeProtocol(nodeURL, protocol);
        try {
            url = UrlBuilder.checkUrl(nodeURL);
            proActiveRuntime = RuntimeFactory.getRuntime(url, protocol);
            jobID = proActiveRuntime.getJobID(url);
        } catch (ProActiveException e) {
            throw new NodeException("Cannot get the node based on " + nodeURL, e);
        } catch (UnknownHostException e) {
            throw new NodeException("Cannot get the node based on " + nodeURL, e);
        }
        Node node = new NodeImpl(proActiveRuntime, url, protocol, jobID);
        return node;
    }

    /**
     * Kills the node of the given url
     * @param nodeURL
     * @throws NodeException if a problem occurs when killing the node
     */
    public static void killNode(String nodeURL) throws NodeException {
        ProActiveRuntime proActiveRuntime;
        String url;

        String protocol = UrlBuilder.getProtocol(nodeURL);
        try {
            url = UrlBuilder.checkUrl(nodeURL);
            proActiveRuntime = RuntimeFactory.getRuntime(url, protocol);
            proActiveRuntime.killNode(url);
        } catch (ProActiveException e) {
            throw new NodeException("Cannot get the node based on " + nodeURL, e);
        } catch (UnknownHostException e) {
            throw new NodeException("Cannot get the node based on " + nodeURL, e);
        }
    }
}
